unit SDBTView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, CommCtrl, DB, autotree;

type
  TSDBTreeView = class;

  {TSDataLinkTreeView}
  TSDataLinkTreeView = class(TDataLink)
  private
    FDBTreeView: TSDBTreeView;
  protected
    procedure ActiveChanged; override;
    procedure DataSetChanged; override;
    procedure DataSetScrolled(Distance: Integer); override;
    procedure RecordChanged(Field: TField); override;
  end;

  TSDBTreeNodes = class;

  {TSDBTreeNode}
  TSDBTreeNode = class
  private
    FDBTreeNodes: TSDBTreeNodes;

    FKeyFieldValue: Variant;
    FParentFieldValue: Variant;
    FTreeNode: TTreeNode;

    Refreshed: Boolean;
    Deleted: Boolean;
    FIndex: Integer;
  protected
    procedure SetKeyFieldValue(Value: Variant);
  public
    constructor Create(AOwner: TSDBTreeNodes);
    property Index: Integer read FIndex;
    property KeyFieldValue: Variant read FKeyFieldValue;
    property ParentFieldValue: Variant read FParentFieldValue;
    property TreeNode: TTreeNode read FTreeNode write FTreeNode;
  end;

  {TSDBTreeNodes}
  TSDBTreeNodes = class
  private
    FlagBuzy: Boolean;
    FDBTreeView: TSDBTreeView;
    FList: TList;
    FMaxKeyFieldValue: Variant;

    function GetDBNodeFromIndex(Index: Integer): TSDBTreeNode;
  protected
    procedure Clear;
    procedure Delete(Index: integer);
    function FindNearest(Value: Variant; var Index: Integer): boolean;
    procedure BeginRefreshRecord;
    procedure RefreshRecord;
    procedure EndRefreshRecord;
    procedure RefreshParents;
    procedure NodeChangeParent(TreeNode: TTreeNode; Value: Variant);

  public
    constructor Create(AOwner: TSDBTreeView);
    destructor Destroy; override;

    function GetDBTreeNode(Value: Variant): TSDBTreeNode;
    function GetTreeNode(Value: Variant): TTreeNode;
    function IndexOf(Value: TTreeNode): integer;
    function Count: Integer;
    function GetParentValue(Value: TTreeNode): Variant;
    function GetKeyFieldValue(Value: TTreeNode): Variant;

    property Items[Index: integer]: TSDBTreeNode read GetDBNodeFromIndex;
    property MaxKeyFieldValue: Variant read FMaxKeyFieldValue;
  end;

  {  }
  TAddNewDBTreeNodeEvent = procedure(Sender: TObject; var DBTreeNode: TSDBTreeNode) of Object;
  TCreateNewKeyValue = procedure(Sender: TObject; var NewKeyValue: Variant) of Object;

  {TSDBTreeView}
  TSDBTreeView = class(TCustomAutoTreeView)
  private
    DataChangedBusy: Boolean;
    NonUserInsertFlag : Boolean;

    FDataLink: TSDataLinkTreeView;
    FDBTreeNodes: TSDBTreeNodes;
    FAddNewItem : TAddNewDBTreeNodeEvent;
    FCreateNewKeyValue : TCreateNewKeyValue;

    FKeyFieldName: string;
    FListFieldName: string;
    FParentFieldName: string;

    FKeyField: TField;
    FListField: TField;
    FParentField: TField;

    procedure DataChanged;
    procedure Scroll;
    procedure RecordEdit(Field: TField);
    procedure RecordInsert(Field: TField);
    procedure RefreshItems;

    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure WndProc(var Message: TMessage); override;

    procedure DataLinkActiveChanged;
    procedure DataLinkRecordChanged(Field: TField);

    function GetSelectedDBTreeNode: TSDBTreeNode;
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure SetKeyFieldName(Value: string);
    procedure SetListFieldName(Value: string);
    procedure SetParentFieldName(Value: string);

    procedure VM_DELETEITEM(var Message: TMessage); message TVM_DELETEITEM;
    procedure VM_INSERTITEM(var Message: TMessage); message TVM_INSERTITEM;
    procedure VM_SELECTITEM(var Message: TMessage); message TVM_SELECTITEM;

    procedure FrameSelectedItem;

  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Change(Node: TTreeNode); override;
    procedure DeleteNodeStructure(List : TList); override;

    procedure Edit(const Item: TTVItem); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure GotoKeyFieldValue(Value: Variant);

    property DBSelected: TSDBTreeNode read GetSelectedDBTreeNode;
    property DBTreeNodes: TSDBTreeNodes read FDBTreeNodes;
    property Items;
    property Selected;

  published
    property Align;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property Images;
    property Indent;
    property KeyField: string read FKeyFieldName write SetKeyFieldName;
    property ListField: string read FListFieldName write SetListFieldName;
    property Options;
    property ParentColor;
    property ParentCtl3D;
    property ParentField: string read FParentFieldName write SetParentFieldName;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property SelectedIndex;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property ShowRoot;
    property SortType;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property Visible;

    property OnAddNewItem : TAddNewDBTreeNodeEvent read FAddNewItem write FAddNewItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnCompare;
    property OnCreateNewKeyValue : TCreateNewKeyValue read FCreateNewKeyValue write FCreateNewKeyValue;
    property OnDblClick;
    property OnDeletion;
    property OnDragDrop;
    property OnDragOver;
    property OnEdited;
    property OnEditing;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnExpanded;
    property OnExpanding;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

var
  BeginEditFlag: Boolean;

procedure Register;

implementation

{____________________________________________________________________}
{Register}

procedure Register;
begin
  RegisterComponents('Standard', [TSDBTreeView]);
end;

{____________________________________________________________________}
{v1  v2 ?}
function VarEquals(const V1, V2: Variant): Boolean;
begin
  Result := False;
  try
    Result := V1 = V2;
  except
  end;
end;

{v1    v2 ?}
function VarFirstMore(const V1, V2: Variant): Boolean;
begin
  Result := False;
  try
    Result := V1 >= V2;
  except
  end;
end;

{____________________________________________________________________}
{TSDataLinkTreeView}

{   }
procedure TSDataLinkTreeView.ActiveChanged;
begin
  if FDBTreeView <> nil then FDBTreeView.DataLinkActiveChanged;
end;

{   }
procedure TSDataLinkTreeView.DataSetChanged;
begin
  FDBTreeView.DataChanged;
end;

{   }
procedure TSDataLinkTreeView.DataSetScrolled(Distance: Integer);
begin
  FDBTreeView.Scroll;
end;

{   }
procedure TSDataLinkTreeView.RecordChanged(Field: TField);
begin
  FDBTreeView.DataLinkRecordChanged(Field);
end;

{____________________________________________________________________}
{TSDBTreeNode}

constructor TSDBTreeNode.Create(AOwner: TSDBTreeNodes);
begin
  FDBTreeNodes := AOwner;
  FKeyFieldValue := NULL;
  FIndex := FDBTreeNodes.FList.Add(Self);
end;

{***************?????????????????????******************}
{  KeyField}
procedure TSDBTreeNode.SetKeyFieldValue(Value: Variant);
var
  newind, i: integer;
begin
  if FKeyFieldValue<>Value then
  begin
    FKeyFieldValue := Value;
    FDBTreeNodes.FindNearest(Value, newind);
    if newind <> FIndex then
    begin
      if newind > FIndex then
        for i := FIndex + 1 to newind do
          Inc(FDBTreeNodes.Items[i].FIndex)
      else
        for i := newind to FIndex - 1 do
          Dec(FDBTreeNodes.Items[i].FIndex);
      FDBTreeNodes.FList.Delete(FIndex);
      FIndex := newind;
      FDBTreeNodes.FList.Insert(FIndex, Self);
    end;
  end;
end;
{***************?????????????????????******************}

{____________________________________________________________________}
{TSDBTreeNodes}

constructor TSDBTreeNodes.Create(AOwner: TSDBTreeView);
begin
  FDBTreeView := AOwner;
  FList := TList.Create;
  FlagBuzy := False;
end;

destructor TSDBTreeNodes.Destroy;
begin
  Clear;
  FList.Free;
  inherited Destroy;
end;

{  }
procedure TSDBTreeNodes.Clear;
var
  i: integer;
begin
  for i := 0 to FList.Count - 1 do
  begin
    TSDBTreeNode(FList.Items[i]).TreeNode.Delete;
    TSDBTreeNode(FList.Items[i]).Free;
  end;
  FList.Clear;
end;

{    Index}
procedure TSDBTreeNodes.Delete(Index: integer);
var
  mDBTreeNode: TSDBTreeNode;
begin
  if (Index >= 0)and(Index < Count) then
  begin
    mDBTreeNode := Items[Index];
    FList.Remove(mDBTreeNode);
    mDBTreeNode.TreeNode.Delete;
    mDBTreeNode.Free;
  end;
end;

{   }
function TSDBTreeNodes.Count: Integer;
begin
  Result := FList.Count;
end;

{   }
function TSDBTreeNodes.GetDBNodeFromIndex(Index: Integer): TSDBTreeNode;
begin
  if (Index > -1)and(Index < Count) then
    Result:=TSDBTreeNode(FList[Index])
  else
    Result:=Nil;
end;

{ DB-,       KeyField  Value}
function TSDBTreeNodes.GetDBTreeNode(Value: Variant): TSDBTreeNode;
var
  i: integer;
begin
  if (FindNearest(Value, i)) then
    Result := Items[i]
  else
    Result := Nil;
end;

{ ,       KeyField  Value}
function TSDBTreeNodes.GetTreeNode(Value: Variant): TTreeNode;
var
  DBTN: TSDBTreeNode;
begin
  DBTN := GetDBTreeNode(Value);
  if DBTN<>nil then
    Result := DBTN.TreeNode
  else
    Result:=Nil;
end;

{   ParentField   }
function TSDBTreeNodes.GetParentValue(Value: TTreeNode): Variant;
var
  i: integer;
begin
  Result := null;
  i := IndexOf(Value);
  if i >= 0 then
    Result := Items[i].ParentFieldValue;
end;

{   KeyField   }
function TSDBTreeNodes.GetKeyFieldValue(Value: TTreeNode): Variant;
var
  i: integer;
begin
  Result := null;
  i := IndexOf(Value);
  if i >= 0 then
    Result := Items[i].KeyFieldValue;
end;

{  ,     Parent}
procedure TSDBTreeNodes.NodeChangeParent(TreeNode: TTreeNode; Value: Variant);
var
  i: integer;
  mTreeNode: TTreeNode;
begin
  if FlagBuzy then exit;
  FlagBuzy := True;
  i := IndexOf(TreeNode);

  { TreeNode  ,  }
  if i < 0 then exit;

  if not VarEquals(Items[i].ParentFieldValue, Value) then
    Items[i].FParentFieldValue := Value;

  {     }
  mTreeNode := GetTreeNode(Value);
  FDBTreeView.MoveTreeNodeStructFlag := True;
  if (mTreeNode <> nil)and(Items[i].TreeNode <> mTreeNode) then
  begin
    TreeNode.Focused := False;
    mTreeNode.HasChildren := True;
    TreeNode.MoveTo(mTreeNode, naAddChild);
  end
  else
    TreeNode.MoveTo(nil, naAdd);
  FDBTreeView.MoveTreeNodeStructFlag := False;
  FDBTreeView.Repaint;
  FlagBuzy := False;
end;

{   Value}
function TSDBTreeNodes.IndexOf(Value: TTreeNode): integer;
var
  i: integer;
begin
  Result := -1;
  if Value <> Nil then
  begin
    for i:=0 to Count -1 do
      if Items[i].TreeNode = Value then
        Break;
      if (i < Count)and(Count > 0) then
        Result:=i;
  end;
end;

{   Value ,   Index   ,   }
                      {Value -    KeyField}
function TSDBTreeNodes.FindNearest(Value: Variant; var Index: Integer): boolean;
var
  Min, Max: Longint;
begin
  Result := False;
  if FList.Count = 0 then
  begin
    Index := -1;
    Exit;
  end;
  Min := 0;
  Max := FList.Count - 1;
  repeat
    if (Max - Min) = 1 then
    begin
      if Min = Index then Min := Max;
      if Max = Index then Max := Min;
    end;
    Index := Min +((Max - Min) div 2);
    if VarEquals(Value, TSDBTreeNode(FList.List^[Index]).FKeyFieldValue) then
      Break;
    if VarFirstMore(Value, TSDBTreeNode(FList.List^[Index]).FKeyFieldValue) then
      Min := Index
    else
      Max := Index;
  until (Min = Max);
  if not VarEquals(Value, TSDBTreeNode(FList.List^[Index]).KeyFieldValue) then
  begin
    if (Index > 0)and VarFirstMore(Value, TSDBTreeNode(FList.List^[Index]).FKeyFieldValue) then
      Dec(Index);
  end
  else
    Result := True;
end;

{  }
procedure TSDBTreeNodes.BeginRefreshRecord;
Var
  i : Integer;
begin
  for i := 0 to Count - 1 do
    Items[i].Refreshed := False;
end;

{ }
procedure TSDBTreeNodes.RefreshRecord;
Var
  mDBTreeNode : TSDBTreeNode;
  mParentTreeNode : TTreeNode;
begin
  FDBTreeView.NonUserInsertFlag := True;

  mDBTreeNode := GetDBTreeNode(FDBTreeView.FKeyField.Value);
  mParentTreeNode := GetTreeNode(FDBTreeView.FParentField.Value);

  {  mDBTreeNode  ,  }
  if(mDBTreeNode = Nil) then begin
     mDBTreeNode := TSDBTreeNode.Create(self);
     mDBTreeNode.SetKeyFieldValue(FDBTreeView.FKeyField.Value);
     mDBTreeNode.FParentFieldValue := FDBTreeView.FParentField.Value;
     mDBTreeNode.FTreeNode := Nil;
  end;

  if Not VarEquals(mDBTreeNode.FKeyFieldValue,  FDBTreeView.FKeyField.Value) then
    mDBTreeNode.SetKeyFieldValue(FDBTreeView.FKeyField.Value);

  if Not VarEquals(mDBTreeNode.FParentFieldValue, FDBTreeView.FParentField.Value) then
    mDBTreeNode.FParentFieldValue := FDBTreeView.FParentField.Value;

  if (mDBTreeNode.TreeNode = Nil) then
  begin
    if(mParentTreeNode = Nil) then
      mDBTreeNode.FTreeNode := FDBTreeView.Items.Add(mParentTreeNode, FDBTreeView.FListField.Text)
    else
      mDBTreeNode.FTreeNode := FDBTreeView.Items.AddChild(mParentTreeNode, FDBTreeView.FListField.Text);

    mDBTreeNode.FTreeNode.Text := FDBTreeView.FListField.Text;

    if Not VarIsNull(mDBTreeNode.FKeyFieldValue) And Assigned(FDBTreeView.FAddNewItem) then
      FDBTreeView.FAddNewItem(Nil, mDBTreeNode);
  end
  else
  begin
    if(mDBTreeNode.TreeNode.Parent <> mParentTreeNode) then
    begin
      mDBTreeNode.TreeNode.MoveTo(mParentTreeNode, naAddChild);
      mParentTreeNode.HasChildren := True;
    end;

    if(mDBTreeNode.TreeNode.Text <> FDBTreeView.FListField.Text) then
      mDBTreeNode.TreeNode.Text := FDBTreeView.FListField.Text;
  end;

  mDBTreeNode.Refreshed := True;
  mDBTreeNode.Deleted := False;
  FDBTreeView.NonUserInsertFlag := False;
end;

{ }
procedure TSDBTreeNodes.RefreshParents;
Var
  i : Integer;
  ParentTreeNode : TTreeNode;
begin
  for i := 0 to Count - 1  do
    if(Items[i].Refreshed) And (Items[i].TreeNode.Parent = Nil) then begin
      ParentTreeNode := GetTreeNode(Items[i].ParentFieldValue);
      FDBTreeView.MoveTreeNodeStructFlag := True;
      if(ParentTreeNode <> Nil) And (Items[i].TreeNode <> ParentTreeNode) then
        Items[i].TreeNode.MoveTo(ParentTreeNode, naAddChild);
      FDBTreeView.MoveTreeNodeStructFlag := False;
    end;
end;

{  }
procedure TSDBTreeNodes.EndRefreshRecord;
Var
 i : Integer;
begin
  i := 0;
  FMaxKeyFieldValue := NULL;
  while i < Count  do begin
    if Not Items[i].Refreshed then begin
       Delete(i);
       i := 0;
    end
    else begin
      if Not VarFirstMore(FMaxKeyFieldValue, Items[i].KeyFieldValue) then
        FMaxKeyFieldValue := Items[i].KeyFieldValue;
      Inc(i);
    end;
  end;
end;


{____________________________________________________________________}
{TSDBTreeView}

constructor TSDBTreeView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataLink := TSDataLinkTreeView.Create;
  FDataLink.FDBTreeView := Self;
  FDBTreeNodes := TSDBTreeNodes.Create(Self);
  DataChangedBusy := False;
end;

destructor TSDBTreeView.Destroy;
begin
  FDBTreeNodes.Clear;
  FDBTreeNodes.Free;
  FDataLink.FDBTreeView := nil;
  FDataLink.Free;
  Items.Clear;
  inherited Destroy;
end;

{  Operation}
procedure TSDBTreeView.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink<>nil) and (AComponent = DataSource) then
    DataSource := nil;
end;

{  Message}
procedure TSDBTreeView.CNNotify(var Message: TWMNotify);
begin
  with Message.NMHdr^ do
    if code = TVN_BEGINLABELEDIT then
      try
        FDataLink.Edit;
        BeginEditFlag := True;
      except
        BeginEditFlag := False;
        raise;
      end;
  inherited;
end;

{    }
procedure TSDBTreeView.DataLinkActiveChanged;
begin
  FKeyField:=Nil;
  FListField:=Nil;
  FParentField:=Nil;
  if FDataLink.Active then
  begin
    if FKeyFieldName<>'' then
      FKeyField:=FDataLink.DataSet.FieldByName(FKeyFieldName);
    if FListFieldName<>'' then
      FListField:=FDataLink.DataSet.FieldByName(FListFieldName);
    if FParentFieldName<>'' then
      FParentField:=FDataLink.DataSet.FieldByName(FParentFieldName);
  end;
    FDBTreeNodes.Clear;
  Items.Clear;
  RefreshItems;
end;

{   }
procedure TSDBTreeView.DataLinkRecordChanged(Field: TField);
begin
  if (Field<>Nil)and(Field.DataSet<>Nil) then
  begin
    case Field.DataSet.State of
      dsEdit: RecordEdit(Field);
      dsInsert: RecordInsert(Field);
    end;
  end;
end;

{  }
procedure TSDBTreeView.RecordEdit(Field: TField);
var
  TreeNode: TTreeNode;
  V: Variant;
begin
  if FKeyField<>nil then
  begin
    TreeNode := Selected;
    if TreeNode = nil then exit;

       { Field -  }
    if (Field = FListField)and(TreeNode.Text<>Field.Text) then
      TreeNode.Text := Field.Text;

       { Field -  Parent}
    if Field = FParentField then
    begin
     {  FDBTreeNodes    TreeNode}
      V := FDBTreeNodes.GetParentValue(TreeNode);

      {        ,
            FDBTreeNode  ,
            ...}
      if not VarIsNull(Field.Value)and not VarIsNull(V)and
        not VarEquals(Field.Value, V) then
          FDBTreeNodes.NodeChangeParent(TreeNode, Field.Value);
    end;
  end;
end;

{  }
procedure TSDBTreeView.RecordInsert(Field : TField);
Var
  mDBTreeNode : TSDBTreeNode;
begin
    if(FKeyField <> Nil) And (FKeyField.Value <> NULL) then begin
      if  Not DataChangedBusy then
        FDBTreeNodes.RefreshRecord;
      mDBTreeNode := DBSelected;
      if(mDBTreeNode <> Nil) And (mDBTreeNode.TreeNode <> Nil) then  begin
        Selected := mDBTreeNode.TreeNode;
        if(Selected <> Nil) then
          Selected.MakeVisible;
     end;
     { }
     if (mDBTreeNode <> Nil) And (mDBTreeNode.ParentFieldValue <> NULL)  And
     Assigned(FAddNewItem) then
       FAddNewItem(Nil, mDBTreeNode);
    end;
end;

{  }
function TSDBTreeView.GetSelectedDBTreeNode: TSDBTreeNode;
begin
  Result:=Nil;
  if (FKeyField<>Nil)and(FKeyField.Value<>NULL)then
    Result:=FDBTreeNodes.GetDBTreeNode(FKeyField.Value);
end;

{  }
function TSDBTreeView.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

{  DataSource}
procedure TSDBTreeView.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource:=Value;
  if Value<>nil then Value.FreeNotification(Self);
end;

{  KeyField}
procedure TSDBTreeView.SetKeyFieldName(Value: string);
begin
  if FKeyFieldName<>Value then
  begin
    FKeyFieldName:=Value;
    DataLinkActiveChanged;
  end;
end;

{  ListField}
procedure TSDBTreeView.SetListFieldName(Value: string);
begin
  if FListFieldName<>Value then
  begin
    FListFieldName:=Value;
    DataLinkActiveChanged;
  end;
end;

{  ParentField}
procedure TSDBTreeView.SetParentFieldName(Value: string);
begin
  if FParentFieldName<>Value then
  begin
    FParentFieldName:=Value;
    DataLinkActiveChanged;
  end;
end;

{ }
procedure TSDBTreeView.DataChanged;
begin
  if DataChangedBusy or not HandleAllocated
    or (FDataLink.DataSet.State = dsInsert) or (FDataLink.DataSet.State = dsEdit) then
    exit;
  RefreshItems;
  Scroll;
end;

{ }
procedure TSDBTreeView.Scroll;
begin
  if (FKeyField <> Nil) then
    GotoKeyFieldValue(FKeyField.Value);
end;

{   Node}
procedure TSDBTreeView.GotoKeyFieldValue(Value: Variant);
var
  TreeNode: TTreeNode;
begin
  TreeNode:=FDBTreeNodes.GetTreeNode(Value);
  if (TreeNode<>Nil) then
  begin
    TreeNode.MakeVisible;
    Selected:=TreeNode;
  end;
end;

{  }
procedure TSDBTreeView.FrameSelectedItem;
Var
  Rect: TRect;
  DC : HDC;
  wnd : HWND;
  OldColor : TColor;
begin
  if Not Focused And (Selected <> Nil) then begin
    TreeView_GetItemRect(Selected.Handle, Selected.ItemId, Rect, True);
    Wnd := handle;
    DC := GetDeviceContext(Wnd);
    OldColor := Brush.Color;
    Brush.Color := Font.Color;
    FrameRect(DC, Rect, Brush.Handle);
    Brush.Color := OldColor;
   end;
end;

{   }
procedure TSDBTreeView.VM_SELECTITEM(var Message: TMessage);
begin
  inherited;
  FrameSelectedItem;
end;

{   }
procedure TSDBTreeView.VM_DELETEITEM(var Message: TMessage);
Var
  mTreeNode : TTreeNode;
  mDBTReeNodeInd : Integer;
begin
 mDBTReeNodeInd := -1;
 if(TVI_ROOT <> HTreeItem(Message.lparam))
 And FDataLink.Active And Not (csDestroying in ComponentState) then begin
   mTreeNode := Items.GetNode(HTreeItem(Message.lparam));
   if(mTreeNode <> Nil)then
     mDBTReeNodeInd := DBTreeNodes.IndexOf(mTreeNode);
 end;
 inherited;
 if (mDBTReeNodeInd >= 0) And DBTreeNodes.Items[mDBTReeNodeInd].Deleted then
   DBTreeNodes.Delete(mDBTReeNodeInd);
end;

{   }
procedure TSDBTreeView.VM_INSERTITEM(var Message: TMessage);
Var
  tn, oldtnp, newtnp : TTreeNode;
  ind : Integer;
  dbtn : TSDBTreeNode;
  str : TTVInsertStruct;
begin
  inherited;
  if Not NonUserInsertFlag  And Not DataChangedBusy
  And Not MoveTreeNodeStructFlag And Not CopyTreeNodeStructFlag then begin
    tn := Items.GetNode(HTreeItem(Message.Result));
    if(tn <> Nil) then
    begin
      ind := DBTreeNodes.IndexOf(tn);

      if(ind < 0) then
        InsertTreeNodeList.Add(tn)
      else
      begin
        dbtn := DBTreeNodes.Items[ind];
        oldtnp := DBTreeNodes.GetTreeNode(dbtn.FParentFieldValue);
        str := PTVInsertStruct(Message.lParam)^;
        newtnp := Items.GetNode(str.hparent);

        if(newtnp <> oldtnp) then
        begin
          DataChangedBusy := True;
          ParentChangedTreeNode := tn;
          NewParentTreeNode := newtnp;
          DataChangedBusy := False;
        end;
      end;
    end;
  end;
end;

{ }
procedure TSDBTreeView.Change(Node: TTreeNode);
Var
  V : Variant;
begin
  inherited;
  if(DataChangedBusy) then exit;
  DataChangedBusy := True;
  if(FKeyField <> Nil) then begin
    V := FDBTreeNodes.GetKeyFieldValue(Node);
    if Not VarIsNull(V) And Not VarEquals(FKeyField.Value, V) And
    (dsInsert <> FKeyField.DataSet.State) And
    Not ((FdataLink.DataSet.EOF) And (FdataLink.DataSet.BOF)) then
      try
        FdataLink.DataSet.Locate(FKeyFieldName, V, []);
      except
      end;
  end;
  DataChangedBusy := False;
end;

{ }
procedure TSDBTreeView.Edit(const Item: TTVItem);
begin
  inherited;
  if(Not ReadOnly) And (FListField <> Nil) And (FListField.DataSet <> Nil) And (Selected <> Nil)
    And ((FListField.DataSet.State = dsEdit) Or (FListField.DataSet.State = dsInsert)) And
    (FListField.Text <> Selected.Text) then
       FListField.Text := Selected.Text
   else Selected.Text := FListField.Text;
end;

{ }
procedure TSDBTreeView.RefreshItems;
Var
  i : Integer;
begin
  DataChangedBusy := True;
  if(FKeyField <> Nil) And (FListField <> Nil) And (FParentField <> Nil) then begin
    {    not Refresh}
    FDBTreeNodes.BeginRefreshRecord;
    { DataSet   }
    FDataLink.DataSet.DisableControls;
    i := 0;
    {   }
    while Not FDataLink.DataSet.EOF do begin
      FDBTreeNodes.RefreshRecord;
      FDataLink.DataSet.Next;
      Inc(i);
    end;
    {   }
    if(i > 0) then
      FDataLink.DataSet.Moveby(-i);
    i := 0;

    {   }
    while Not FDataLink.DataSet.BOF do begin
      FDBTreeNodes.RefreshRecord;
      FDataLink.DataSet.Prior;
      Inc(i);
    end;

    {   }
    if(i > 0) then
      FDataLink.DataSet.Moveby(i);
    { DataSet}
    FDataLink.DataSet.EnableControls;

    FDBTreeNodes.RefreshParents;
    { }
    FDBTreeNodes.EndRefreshRecord;
  end;

  DataChangedBusy := False;
end;

{  }
procedure TSDBTreeView.DeleteNodeStructure(List : TList);
Var
  i : Integer;
  mDBTReeNodeInd : Integer;
  flag : Boolean;
begin
  if Not FDataLink.Active  then exit;
  FDataLink.DataSet.DisableControls;
  for i := List.Count - 1 downto 0 do begin
    mDBTReeNodeInd := DBTreeNodes.IndexOf(TTreeNode(List.List^[i]));
     if (mDBTReeNodeInd >= 0) then begin
       flag := False;
       if  Not ((FdataLink.DataSet.EOF) And (FdataLink.DataSet.BOF)) then
         try
           flag := FdataLink.DataSet.Locate(FKeyFieldName, DBTreeNodes.Items[mDBTReeNodeInd].KeyFieldValue, []);
         except
         end;
       If flag And VarEquals(DBTreeNodes.Items[mDBTReeNodeInd].KeyFieldValue, FKeyField.Value) then
         FDataLink.DataSet.Delete;
       DBTreeNodes.Items[mDBTReeNodeInd].Deleted := True;
     end;
  end;
  flag := DataChangedBusy;
  DataChangedBusy := True;
  FDataLink.DataSet.EnableControls;
  DataChangedBusy := flag;
end;

procedure TSDBTreeView.WndProc(var Message: TMessage);
begin
  inherited WndProc(Message);
end;

{____________________________________________________________________}

initialization
  BeginEditFlag := True;

end.
